@@ -1,5 +1,5 @@
 ### =========================================================================
-### Helper functions not exported 
+### Helper functions not exported
 ### =========================================================================
 
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -8,7 +8,7 @@
 
 .collapseLists <- function(vcf, param)
 {
-    idx <- sapply(vcf, function(elt) length(elt$rowData) > 0L)
+    idx <- sapply(vcf, function(elt) length(elt$rowRanges) > 0L)
     if (!sum(idx))
         return(vcf[[1]])
     if (length(vcf) > 1L)
@@ -23,55 +23,55 @@
     ## single range in 'which'
     if (1L == length(vcf)) {
         lst <- vcf[[1]]
-        lst$paramRangeID <- as.factor(rep(paramRangeID, length(lst$rowData)))
+        lst$paramRangeID <- as.factor(rep(paramRangeID, length(lst$rowRanges)))
     } else {
     ## multiple ranges in 'which'
         lst <- lapply(names(vcf[[1]]), function(elt) {
                    suppressWarnings(do.call(c, unname(lapply(vcf, "[[", elt))))
                })
         names(lst) <- names(vcf[[1]])
-        len <- unlist(lapply(vcf, function(elt) length(elt$rowData)),
+        len <- unlist(lapply(vcf, function(elt) length(elt$rowRanges)),
                       use.names=FALSE)
-        paramRangeID <- as.factor(rep(paramRangeID, len)) 
+        paramRangeID <- as.factor(rep(paramRangeID, len))
 
         ## collapse info and geno
         info <- lst$INFO
         sp <- split(unname(info), unique(names(info)))
-        sp <- sp[unique(names(info))] 
-        lst$INFO <- 
+        sp <- sp[unique(names(info))]
+        lst$INFO <-
             lapply(sp, function(elt) {
                 d <- dim(elt[[1]])
                 if (is(elt[[1]], "list")) {
                     as.matrix(elt)
                 } else if (is(elt[[1]], "array") && !is.na(d[3])) {
                     pc <- lapply(seq_len(d[2]), function(i) {
-                              do.call(rbind, lapply(elt, "[", ,i,)) 
+                              do.call(rbind, lapply(elt, "[", ,i,))
                           })
-                    array(do.call(c, pc), 
-                        c(length(lst$rowData), d[2], d[3]))
+                    array(do.call(c, pc),
+                        c(length(lst$rowRanges), d[2], d[3]))
                 } else {
                     do.call(c, elt)
                 }
-            }) 
+            })
         geno <- lst$GENO
         sp <- split(geno, unique(names(geno)))
-        lst$GENO <- 
+        lst$GENO <-
             lapply(sp, function(elt) {
                 d <- dim(elt[[1]])
                 if (!is.na(d[3])) {
                     pc <- lapply(seq_len(d[3]), function(i) {
-                              do.call(rbind, lapply(elt, "[", ,,i)) 
+                              do.call(rbind, lapply(elt, "[", ,,i))
                           })
-                    cmb <- array(do.call(c, pc), 
-                                 c(length(lst$rowData), d[2], d[3]))
+                    cmb <- array(do.call(c, pc),
+                                 c(length(lst$rowRanges), d[2], d[3]))
                     cmb
                 } else {
                     trans <- lapply(elt, t)
-                    cmb <- matrix(do.call(c, trans), length(lst$rowData), 
+                    cmb <- matrix(do.call(c, trans), length(lst$rowRanges),
                                   d[2], byrow=TRUE)
                     cmb
                 }
-            }) 
+            })
         lst$paramRangeID <- paramRangeID
     }
     lst
@@ -79,19 +79,20 @@
 
 .formatList <- function(data, type)
 {
-    switch(type, 
+    switch(type,
         Integer = IntegerList(data),
         Float = NumericList(data),
         String = CharacterList(data),
+        Character = CharacterList(data),
         Logical = LogicalList(data))
 }
 
 .toDNAStringSetList <- function(x)
 {
-    ### also used in predictCoding(), genotypeToSnpMatrix() 
+    ### also used in predictCoding(), genotypeToSnpMatrix()
     pbw <- PartitioningByWidth(elementLengths(x))
     x <- unlist(x, use.names=FALSE)
-    x[.isStructural(x)] <- "" 
+    x[.isStructural(x)] <- ""
     xx <- sub(".", "", x, fixed=TRUE)
     relist(DNAStringSet(xx), pbw)
 }
@@ -115,18 +116,17 @@
 .formatInfo <- function(x, hdr, nrecords)
 {
     ## no data
-    if (0L == length(x) ||
-       (1L == length(x) && all(names(x) == "INFO")))
+    if (!length(x) || (1L == length(x) && all(x[[1]] == ".", na.rm=TRUE)))
         return(DataFrame(row.names=seq_len(nrecords)))
     ## data in file but not in header
-    if (0L == length(hdr)) {
+    if (!length(hdr) && length(x[[1]])) {
         DF <- DataFrame(x)
         names(DF) <- names(x)
         return(DF)
     }
     ## matrices and arrays
     type <- hdr$Type[match(names(x), rownames(hdr))]
-    idx <- which(lapply(x, is.array) == TRUE) 
+    idx <- which(lapply(x, is.array) == TRUE)
     if (0L != length(idx)) {
         for (i in idx) {
             dat <- x[[i]]
@@ -138,7 +138,7 @@
             x[[i]] <- .formatList(dat, type[i])
         }
     }
-    ## ragged lists 
+    ## ragged lists
     lx <- which(lapply(x, is.list) == TRUE)
     if (0L != length(lx)) {
         for (i in lx) {
@@ -150,28 +150,22 @@
     DF
 }
 
-
-.listCumsum <- function(x) {
-  x_unlisted <- unlist(x, use.names=FALSE)
-  x_cumsum <- cumsum(x_unlisted)
-  x_part <- PartitioningByWidth(elementLengths(x))
-  x_cumsum - rep(x_cumsum[start(x_part)] - x_unlisted[start(x_part)],
-                 width(x_part))
-}
-
-.listCumsumShifted <- function(x) {
-  cs <- .listCumsum(x)
-  shifted <- c(0L, head(cs, -1))
-  shifted[start(PartitioningByWidth(elementLengths(x)))] <- 0L
-  shifted
-}
-
 .rleRecycleVector <- function(x, len) {
   if (is(x, "Rle"))
     rep(x, length.out = len)
   else if (length(x) == 1L)
     Rle(x, len)
-  else IRanges:::recycleVector(x, len)
+  else S4Vectors:::recycleVector(x, len)
+}
+
+.pasteCollapseRows <- function(x, sep = ",") {
+  if (!is.matrix(x) || mode(x) != "character") {
+    stop("'x' must be a matrix of mode character")
+  }
+  if (!isSingleString(sep) || nchar(sep) == 0L) {
+    stop("'sep' must be a single, non-NA, non-empty string")
+  }
+  .Call(matrix_pasteCollapseRows, x, sep)
 }
 
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -189,7 +183,7 @@
     if (0L == length(vals))
         return("")
     sql <-
-      lapply(seq_len(length(vals)), 
+      lapply(seq_len(length(vals)),
         function(i) {
             v <- vals[[i]]
             if (!is.numeric(v))
@@ -207,7 +201,7 @@
         msg <- paste(BiocGenerics:::selectSome(keys[mkeys]), collapse=" ")
         warning(sum(mkeys), " keys not found in ", db, " database: ", msg,
                 call.=FALSE)
-    } 
+    }
     all(mkeys)
 }
 
@@ -222,11 +216,11 @@
         return(TRUE)
     } else {
         return(FALSE)
-    } 
+    }
 }
 
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-### isSNV helpers 
+### isSNV helpers
 ###
 
 .isSNV <- function(ref, alt) {
@@ -234,21 +228,23 @@
 }
 
 .isDeletion <- function(ref, alt) {
-    nchar(alt) == 1L & 
-    nchar(ref) > 1L & 
-    substring(ref, 1, 1) == alt
+    nchar(alt) == 1L & nchar(ref) > 1L & substring(ref, 1, 1) == alt
 }
 
 .isInsertion <- function(ref, alt) {
     nchar(ref) == 1L & 
     nchar(alt) > 1L & 
-    substring(alt, 1, 1) == ref
+    substring(alt, 1, 1) == as.character(ref)
 }
 
 .isIndel <- function(ref, alt) {
     .isDeletion(ref, alt) | .isInsertion(ref, alt)
 }
 
+.isDelins <- function(ref, alt) {
+    !.isIndel(ref, alt) & !.isSubstitution(ref, alt)
+}
+
 .isTransition <- function(ref, alt) {
     m1 <- match(as.character(ref), c("A", "G", "T", "C"))
     m2 <- match(as.character(alt), c("G", "A", "C", "T"))
@@ -260,3 +256,51 @@
 .isSubstitution <- function(ref, alt) {
     nchar(ref) == nchar(alt)
 }
+
+### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+### predictCoding()
+###
+
+.localCoordinates <- function(from, to, ignore.strand, ...)
+{
+    ## 'to' is a GRangesList of cds by transcript
+    map <- mapToTranscripts(unname(from), to, ignore.strand=ignore.strand, ...)
+    if (length(map) == 0) {
+        res <- GRanges()
+        mcols(res) <- DataFrame(REF=DNAStringSet(), ALT=DNAStringSetList(),
+                                varAllele=DNAStringSet(), CDSLOC=IRanges(),
+                                PROTEINLOC=IntegerList(), QUERYID=integer(),
+                                TXID=character(), CDSID=IntegerList())
+        return(res)
+    }
+
+    xHits <- map$xHits
+    txHits <- map$transcriptsHits
+    flat_to <- unlist(to) ## names needed for mapping
+
+    ## FIXME: cdsid is expensive
+    cdsid <- IntegerList(integer(0))
+    map2 <- mapToTranscripts(unname(from)[xHits], flat_to,
+                             ignore.strand=ignore.strand)
+    cds <- mcols(flat_to)$cds_id[map2$transcriptsHits]
+    if (length(cds)) {
+        cdslst <- unique(splitAsList(cds, map2$xHits))
+        cdsid <- cdslst
+    }
+    if (is.null(txid <- names(to)[txHits]))
+        txid <- NA_integer_
+
+    ## protein coordinates
+    pends <- c(ceiling(start(map)/3), ceiling(end(map)/3))
+    plocs <- unique(IntegerList(split(pends, rep(seq_len(length(pends)/2)), 2)))
+
+    res <- from[xHits]
+    strand(res) <- strand(map)
+    mcols(res) <- append(values(res), 
+        DataFrame(CDSLOC=ranges(map), 
+                  PROTEINLOC=plocs, 
+                  QUERYID=xHits, 
+                  TXID=txid, CDSID=cdsid))
+    res 
+}
+
