@@ -8,10 +8,10 @@
 ###
 
 VCF <-
-    function(rowData=GRanges(), colData=DataFrame(), 
+    function(rowRanges=GRanges(), colData=DataFrame(), 
              exptData=SimpleList(header=VCFHeader()), 
              fixed=DataFrame(),
-             info=DataFrame(row.names=seq_along(rowData)), 
+             info=DataFrame(row.names=seq_along(rowRanges)), 
              geno=SimpleList(),
              ..., collapsed=TRUE, verbose=FALSE)
 {
@@ -20,23 +20,23 @@ VCF <-
         class <- "CollapsedVCF"
         if (!length(fixed)) {
             fixed=DataFrame(
-                REF=DNAStringSet(rep(".", length(rowData))), 
-                ALT=DNAStringSetList(as.list(rep(".", length(rowData)))),
-                QUAL=rep(NA_real_, length(rowData)), 
-                FILTER=rep(NA_character_, length(rowData)))
+                REF=DNAStringSet(rep(".", length(rowRanges))), 
+                ALT=DNAStringSetList(as.list(rep(".", length(rowRanges)))),
+                QUAL=rep(NA_real_, length(rowRanges)), 
+                FILTER=rep(NA_character_, length(rowRanges)))
         }
     } else {
         class <- "ExpandedVCF"
         if (!length(fixed)) {
             fixed=DataFrame(
-                REF=DNAStringSet(rep("", length(rowData))), 
-                ALT=DNAStringSet(rep("", length(rowData))),
-                QUAL=rep(NA_real_, length(rowData)), 
-                FILTER=rep(NA_character_, length(rowData)))
+                REF=DNAStringSet(rep("", length(rowRanges))), 
+                ALT=DNAStringSet(rep("", length(rowRanges))),
+                QUAL=rep(NA_real_, length(rowRanges)),
+                FILTER=rep(NA_character_, length(rowRanges)))
         }
     }
 
-    new(class, SummarizedExperiment(assays=geno, rowData=rowData,
+    new(class, SummarizedExperiment(assays=geno, rowRanges=rowRanges,
         colData=colData, exptData=exptData), fixed=fixed, info=info, ...)
 }
 
@@ -47,7 +47,7 @@ VCF <-
 .checkLength <- function (x, len) 
 {
     if (len != length(slot(x, "rowData"))) 
-        stop("length(value) must equal length(rowData(x))")
+        stop("length(value) must equal length(rowRanges(x))")
 }
 
 ### ref 
@@ -133,30 +133,30 @@ setReplaceMethod("fixed", c("VCF", "DataFrame"),
     x
 })
 
-### rowData
-setMethod("rowData", "VCF", 
-    function(x) 
+### rowRanges
+setMethod("rowRanges", "VCF", 
+    function(x, ..., fixed = TRUE) 
 {
-    gr <- slot(x, "rowData") 
-    if (length(slot(x, "fixed")) != 0L)
-        mcols(gr) <- append(mcols(gr), slot(x, "fixed"))
+    gr <- slot(x, "rowData")
+    if (fixed)
+        if (length(slot(x, "fixed")) != 0L)
+            mcols(gr) <- append(mcols(gr), slot(x, "fixed"))
     gr
 })
 
-setReplaceMethod("rowData", c("VCF", "GRanges"),
+setReplaceMethod("rowRanges", c("VCF", "GRanges"),
     function(x, value)
 {
-    idx <- names(mcols(value)) %in% "paramRangeID"
-    slot(x, "fixed") <- mcols(value)[!idx] 
-    slot(x, "rowData") <- value[,idx]
+    fixed_idx <- names(mcols(value)) %in% c("REF", "ALT", "QUAL", "FILTER")
+    slot(x, "fixed") <- mcols(value)[fixed_idx] 
+    slot(x, "rowData") <- value[,!fixed_idx]
     validObject(x)
     x
 })
 
-### Must define VCF methods for 'mcols<-' and 'dimnames<-'
-### instead of inheriting from SummarizedExperiment.
-### The 'fixed' fields are stored in a separtate slot
-### and not as metadata columns of 'rowData'.
+### Must define VCF methods for 'mcols<-' and 'dimnames<-' instead of 
+### inheriting from SummarizedExperiment. 'fixed' fields are stored in 
+### a separate slot and not as metadata columns of 'rowRanges'.
 
 setReplaceMethod("mcols", c("VCF", "DataFrame"),
     function(x, ..., value)
@@ -170,21 +170,20 @@ setReplaceMethod("mcols", c("VCF", "DataFrame"),
 setReplaceMethod("dimnames", c("VCF", "list"),
     function(x, value)
 {
-    rowData <- slot(x,"rowData")
-    names(rowData) <- value[[1]]
+    rowRanges <- slot(x,"rowData")
+    names(rowRanges) <- value[[1]]
     colData <- colData(x)
     rownames(colData) <- value[[2]]
-    GenomicRanges:::clone(x, rowData=rowData, colData=colData)
+    GenomicRanges:::clone(x, rowData=rowRanges, colData=colData)
 })
  
 ### info 
 setMethod("info", "VCF", 
-    function(x) 
+    function(x, ..., row.names = TRUE) 
 {
     info <- slot(x, "info")
-    if (length(info) != 0L)
-        if (!sum(duplicated(rownames(x))))
-            rownames(info) <- rownames(x)
+    if (row.names && !anyDuplicated(rownames(x)))
+        rownames(info) <- rownames(x)
     info
 })
 
@@ -256,13 +255,13 @@ setReplaceMethod("geno", c("VCF", "missing", "matrix"),
 setMethod("strand", "VCF",
     function(x, ...)
 {
-    strand(rowData(x))
+    strand(rowRanges(x))
 })
 
 setReplaceMethod("strand", "VCF",
     function(x, ..., value)
 {
-    strand(rowData(x)) <- value
+    strand(rowRanges(x)) <- value
     x
 })
 
@@ -372,11 +371,21 @@ setMethod("rbind", "VCF",
     args <- unname(list(...))
     if (!.compare(lapply(args, class)))
         stop("'...' objects must be of the same VCF class")
-    args <- .renameSamples(args) 
-    se <- GenomicRanges:::.rbind.SummarizedExperiment(args)
+    args <- .renameSamples(args)
+    if (!.compare(lapply(args, colnames)))
+            stop("'...' objects must have the same colnames")
+    if (!.compare(lapply(args, ncol)))
+            stop("'...' objects must have the same number of samples")
+
+    rowRanges <- do.call(c, lapply(args,
+        function(i) slot(i, "rowData")))
+    colData <- GenomicRanges:::.cbind.DataFrame(args, colData, "colData")
+    assays <- GenomicRanges:::.bind.arrays(args, rbind, "assays")
+    exptData <- do.call(c, lapply(args, exptData))
     info <- do.call(rbind, lapply(args, info))
     fixed <- do.call(rbind, lapply(args, fixed))
-    new(class(args[[1]]), se, fixed=fixed, info=info)
+    new(class(args[[1]]), assays=assays, rowData=rowRanges,
+        colData=colData, exptData=exptData, fixed=fixed, info=info)
 })
 
 .renameSamples <- function(args)
@@ -388,6 +397,7 @@ setMethod("rbind", "VCF",
     }, x=args, i=seq_along(args)) 
 }
 
+## FIXME: combine header info
 ## Appropriate for objects with same ranges and different samples.
 setMethod("cbind", "VCF",
     function(..., deparse.level=1)
@@ -395,13 +405,24 @@ setMethod("cbind", "VCF",
     args <- unname(list(...))
     if (!.compare(lapply(args, class)))
         stop("'...' objects must be of the same VCF class")
-    se <- GenomicRanges:::.cbind.SummarizedExperiment(args)
+
+
+    if (!.compare(lapply(args, rowRanges), TRUE))
+        stop("'...' object ranges (rows) are not compatible")
+    rowRanges <- rowRanges(args[[1]])
+    mcols(rowRanges) <- GenomicRanges:::.cbind.DataFrame(args, mcols, "mcols")
     info <- GenomicRanges:::.cbind.DataFrame(args, info, "info") 
+    colData <- do.call(rbind, lapply(args, colData))
+    assays <- GenomicRanges:::.bind.arrays(args, cbind, "assays")
+    exptData <- do.call(c, lapply(args, exptData))
     if (!.compare(lapply(args, "fixed")))
         stop("data in 'fixed(VCF)' must match.")
     else
         fixed <- fixed(args[[1]]) 
-    new(class(args[[1]]), se, fixed=fixed, info=info)
+    new(class(args[[1]]), assays=assays, rowData=rowRanges,
+        colData=colData, exptData=exptData, fixed=fixed, info=info)
+
+
 })
 
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -418,18 +439,20 @@ setMethod(show, "VCF",
 ### methods for CollapsedVCF and ExapandedVCF
 .showVCFSubclass <- function(object)
 {
-    prettydescr <- function(desc) {
+    prettydescr <- function(desc, wd0=30L) {
         desc <- as.character(desc)
-        wd <- options()[["width"]] - 30
+        wd <- options()[["width"]] - wd0
         dwd <- nchar(desc)
         desc <- substr(desc, 1, wd)
         idx <- dwd > wd
         substr(desc[idx], wd - 2, dwd[idx]) <- "..."
         desc
     }
-    headerrec <- function(df, lbl, margin="  ") {
-        df$Description <- prettydescr(df$Description)
-        rownames(df) <- paste(margin, rownames(df))
+    headerrec <- function(df, margin="   ") {
+        wd <- max(nchar(rownames(df))) + nchar("Number") +
+            max(nchar(df$Type)) + 7L
+        df$Description <- prettydescr(df$Description, wd)
+        rownames(df) <- paste0(margin, rownames(df))
         print(df, right=FALSE)
     }
     printSmallGRanges <- function(x, margin)
@@ -459,17 +482,19 @@ setMethod(show, "VCF",
     margin <- "  "
     cat("class:", class(object), "\n")
     cat("dim:", dim(object), "\n")
-    cat("rowData(vcf):\n")
-    printSmallGRanges(rowData(object), margin=margin)
+    cat("rowRanges(vcf):\n")
+    printSmallGRanges(rowRanges(object), margin=margin)
     cat("info(vcf):\n")
-    printSmallDataTable(info(object), margin=margin) 
+    printSmallDataTable(info(object, row.names=FALSE), margin=margin) 
     if (length(header(object))) {
         if (length(hdr <- info(header(object)))) {
-            nms <- intersect(colnames(info(object)), rownames(hdr))
-            diff <- setdiff(colnames(info(object)), rownames(hdr))
+            nms <- intersect(colnames(info(object, row.names=FALSE)),
+                             rownames(hdr))
+            diff <- setdiff(colnames(info(object, row.names=FALSE)),
+                            rownames(hdr))
             if (length(nms)) {
                 cat("info(header(vcf)):\n")
-                headerrec(as.data.frame(hdr[nms,]), "info")
+                headerrec(as.data.frame(hdr[nms,]))
             }
             if (length(diff))
                 cat("  Fields with no header:", 
@@ -485,7 +510,7 @@ setMethod(show, "VCF",
             diff <- setdiff(names(geno(object)), rownames(hdr))
             if (length(nms)) {
                 cat("geno(header(vcf)):\n")
-                headerrec(as.data.frame(hdr[nms,]), "geno")
+                headerrec(as.data.frame(hdr[nms,]))
             }
             if (length(diff))
                 cat("  Fields with no header:", 
@@ -504,7 +529,7 @@ setMethod("updateObject", "VCF",
     {
         if (verbose) 
             message("updateObject(object = 'VCF')") 
-        VCF(rowData=rowData(object), colData=colData(object), exptData=exptData(object), 
+        VCF(rowRanges=rowRanges(object), colData=colData(object), exptData=exptData(object), 
             info=mcols(info(object))[-1], fixed=mcols(fixed(object))[-1], geno=geno(object))
     }
 )
@@ -517,18 +542,24 @@ setMethod("updateObject", "VCF",
 ### or ExpandedVCF. Data in 'alt(x)' must be DNAStringSet or DNAStringSetList.
 restrictToSNV <- function(x, ...)
 {
-    if (!is(x, "VCF"))
-        stop("'x' must be a VCF object")
-    if (is(alt <- alt(x), "CharacterList"))
-        stop("'alt' must be non-structural nucleotide values")
-
-    ref_snp <- nchar(ref(x)) == 1
-    if (is(alt, "DNAStringSetList"))
-        alt_snp <- any(relist((nchar(unlist(alt)) == 1), alt))
-    else if (is(alt, "DNAStringSet"))
-        alt_snp <- nchar(alt) == 1
-    else
-        stop("unrecognized 'alt' data type")
-
-    x[ref_snp & alt_snp]
+    .Deprecated("isSNV")
 }
+
+### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+### genotypeCodesToNucleotides
+###
+
+genotypeCodesToNucleotides <- function(vcf, ...) 
+{
+    GT <- geno(vcf, withDimnames=FALSE)$GT
+    if (is.null(GT <- geno(vcf, withDimnames=FALSE)$GT))
+        stop("no 'GT' data found in geno(vcf)")
+
+    if (is(alt(vcf), "CharacterList"))
+        stop("'ALT' must be a DNAStringSetList")
+    ALT <- split(as.character(alt(vcf)@unlistData), togroup(alt(vcf))) 
+    REF <- as.character(ref(vcf))
+    geno(vcf)$GT <- .geno2geno(NULL, ALT, REF, GT)
+    vcf
+}
+
