@@ -3,6 +3,15 @@
 ### -------------------------------------------------------------------------
 ###
 
+### Thoughts on the gVCF format.
+## The gVCF format is essentially a run-length encoding of wildtype
+## calls. We could represent the run as a range, but we would need
+## somewhat to indicate "WT". Probably best to go the VCF route: have
+## the first ref base, and an NA alt. We can distinguish this from an
+## SNV or indel, because the width of the range is > 1, but there is
+## only a single ref base. This will break the assertion that the ref
+## length always equal the range width, but that is not a big deal.
+
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 ### Accessors
 ###
@@ -56,32 +65,36 @@ setMethod("altFraction", "VRanges", function(x) {
   altDepth(x) / totalDepth(x)
 })
 
+setMethod("refFraction", "VRanges", function(x) {
+    refDepth(x) / totalDepth(x)
+})
+
 setMethod("isDeletion", "VRanges", function(x, ...) 
-  .isDeletion(ref(x), alt(x))
+    .dispatchSNV_ExpandedVCF(.isDeletion, x)
 )
-  #nchar(alt(x)) == 1L & nchar(ref(x)) > 1L & substring(ref(x), 1, 1) == alt(x)
 
 setMethod("isInsertion", "VRanges", function(x, ...) 
-    .isInsertion(ref(x), alt(x)) 
+    .dispatchSNV_ExpandedVCF(.isInsertion, x)
 )
- # nchar(ref(x)) == 1L & nchar(alt(x)) > 1L & substring(alt(x), 1, 1) == ref(x)
 
 setMethod("isIndel", "VRanges", function(x, ...)
-    .isIndel(ref(x), alt(x)) 
+    .dispatchSNV_ExpandedVCF(.isIndel, x)
+)
+
+setMethod("isDelins", "VRanges", function(x, ...)
+    .dispatchSNV_ExpandedVCF(.isDelins, x)
 )
- # isDeletion(x) | isInsertion(x)
 
 setMethod("isSNV", "VRanges", function(x, ...)
-    .isSNV(ref(x), alt(x)) 
+    .dispatchSNV_ExpandedVCF(.isSNV, x)
 )
- # nchar(alt(x)) == 1L & nchar(ref(x)) == 1L
 
 setMethod("isSubstitution", "VRanges", function(x, ...)
-    .isSubstitution(ref(x), alt(x)) 
+    .dispatchSNV_ExpandedVCF(.isSubstitution, x)
 )
 
 setMethod("isTransition", "VRanges", function(x, ...)
-    .isTransition(ref(x), alt(x)) 
+    .dispatchSNV_ExpandedVCF(.isTransition, x)
 )
 
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -109,7 +122,7 @@ VRanges <-
   if (length(gr) != maxLen)
     gr <- rep(gr, length.out = maxLen)
   ref <- as.character(ref)
-  ref <- IRanges:::recycleVector(ref, maxLen)
+  ref <- S4Vectors:::recycleVector(ref, maxLen)
   alt <- .rleRecycleVector(alt, maxLen)
   alt <- as(alt, "characterOrRle")
   refDepth <- .rleRecycleVector(refDepth, maxLen)
@@ -118,7 +131,7 @@ VRanges <-
   softFilterMatrix <- as.matrix(softFilterMatrix)
   mode(softFilterMatrix) <- "logical"
   softFilterMatrix.ind <-
-    IRanges:::recycleVector(seq_len(nrow(softFilterMatrix)), maxLen)
+    S4Vectors:::recycleVector(seq_len(nrow(softFilterMatrix)), maxLen)
   softFilterMatrix <- softFilterMatrix[softFilterMatrix.ind,,drop=FALSE]
   sampleNames <- .rleRecycleVector(sampleNames, maxLen)
   totalDepth <- as(totalDepth, "integerOrRle")
@@ -139,6 +152,8 @@ VRanges <-
 ### Coercion
 ###
 
+setMethod("relistToClass", "VRanges", function(x) "CompressedVRangesList")
+
 parseFilterStrings <- function(x) {
   x[x == "."] <- NA
   if (all(is.na(x)))
@@ -173,9 +188,9 @@ genoToMCol <- function(x) {
 
 setAs("VCF", "VRanges", function(from) {
   from <- expand(from)
-  rd <- rowData(from)
+  rd <- rowRanges(from)
   seqnames <- seqnames(rd)
-  ranges <- unname(ranges(rd))
+  ranges <- ranges(rd)
   ref <- rd$REF
   alt <- rd$ALT
   if (is(alt, "DNAStringSetList") || is(alt, "CharacterList"))
@@ -203,8 +218,13 @@ setAs("VCF", "VRanges", function(from) {
     }
   } else sampleNames <- NA_character_
   meta <- info(from)
-  if (!is.null(mcols(rd)$QUAL))
+  if (!is.null(mcols(rd)$QUAL)) {
     meta <- cbind(mcols(rd)["QUAL"], meta)
+  }
+  if (!is.null(meta$END)) {
+    end(ranges)[!is.na(meta$END)] <- meta$END[!is.na(meta$END)]
+    meta$END <- NULL
+  }
   rownames(meta) <- NULL
   if (!is.null(rd$FILTER))
     filter <- parseFilterStrings(rd$FILTER)
@@ -234,6 +254,14 @@ setAs("VRanges", "VCF", function(from) {
   asVCF(from)
 })
 
+optionalDescriptions <- c(
+  GT = "Genotype",
+  GQ = "Genotype quality",
+  PL = "Normalized, Phred-scaled likelihoods for genotypes",
+  MIN_DP = "Minimum DP observed within the gVCF block",
+  END = "End position of the gVCF WT run"
+  )
+  
 makeFORMATheader <- function(x) {
   fixed <-
     DataFrame(row.names = c("AD", "DP", "FT"),
@@ -275,7 +303,7 @@ makeINFOheader <- function(x) {
   df$Number <- as.character(lapply(x, numberForColumn))
   df$Type <- as.character(lapply(x, typeForColumn))
   if (is.null(df$Description))
-    df$Description <- rep(NA, nrow(df))
+    df$Description <- optionalDescriptions[rownames(df)]
   df[c("Number", "Type", "Description")]
 }
 
@@ -311,10 +339,18 @@ makeFILTERstrings <- function(x) {
   failedRows <- rowSums(!x, na.rm=TRUE) > 0L
   ftStrings[passedRows & !failedRows] <- "PASS"
   ftStrings[failedRows] <-
-    unstrsplit(seqsplit(ftNames, row(x)[failed]), ";")
+    unstrsplit(splitAsList(ftNames, row(x)[failed]), ";")
   ftStrings
 }
 
+isGVCFRun <- function(x) {
+  nchar(ref(x)) == 1L & width(x) > 1L
+}
+
+gVCFRunEnds <- function(x) {
+  ifelse(isGVCFRun(x), end(x), NA_integer_)
+}
+
 vranges2Vcf <- function(x, info = character(), filter = character(),
                         meta = character())
 {
@@ -345,9 +381,18 @@ vranges2Vcf <- function(x, info = character(), filter = character(),
     xUniq <- x
   }
 
-  rowData <- xUniq
-  mcols(rowData) <- NULL
-  rowData <- as(rowData, "GRanges", strict=TRUE)
+  END <- gVCFRunEnds(xUniq)
+  if (!all(is.na(END))) {
+    if (!is.null(xUniq$END)) {
+      warning("Replacing 'END' metadata/info column with computed gVCF END")
+    }
+    xUniq$END <- END
+    info <- union(info, "END")
+  }
+  
+  rowRanges <- xUniq
+  mcols(rowRanges) <- NULL
+  rowRanges <- as(rowRanges, "GRanges", strict=TRUE)
 
   colData <- DataFrame(Samples = seq_along(sampleLevels),
                        row.names = sampleLevels)
@@ -362,7 +407,7 @@ vranges2Vcf <- function(x, info = character(), filter = character(),
   header <- VCFHeader(reference = seqlevels(x), samples = sampleLevels,
                       header = DataFrameList(META = meta_header,
                         FORMAT = makeFORMATheader(mcols(x)[genoMCols]),
-                        INFO = makeINFOheader(mcols(x)[info]),
+                        INFO = makeINFOheader(mcols(xUniq)[info]),
                         FILTER = makeFILTERheader(x)))
   exptData <- SimpleList(header = header)
 
@@ -428,8 +473,10 @@ vranges2Vcf <- function(x, info = character(), filter = character(),
   if (length(genoMCols) > 0L)
     geno <- c(geno, SimpleList(lapply(mcols(x)[genoMCols], genoArray)))
 
-  VCF(rowData = rowData, colData = colData, exptData = exptData, fixed = fixed,
-      geno = geno, info = mcols(xUniq)[info], collapsed = FALSE)
+  info <- mcols(xUniq)[info]
+  
+  VCF(rowRanges = rowRanges, colData = colData, exptData = exptData,
+      fixed = fixed, geno = geno, info = info, collapsed = FALSE)
 }
 
 setMethod("asVCF", "VRanges", vranges2Vcf)
@@ -439,17 +486,35 @@ setMethod("asVCF", "VRanges", vranges2Vcf)
 ###
 
 VRangesScanVcfParam <- function(fixed="ALT", info=NA, geno="AD", ...) {
+  .Deprecated("ScanVcfParam")
   ScanVcfParam(fixed=fixed, info=info, geno=geno, ...)
 }
 
-readVcfAsVRanges <- function(x, genome, param=VRangesScanVcfParam(), ...) {
-  as(readVcf(x, genome, param=param, ...), "VRanges")
+readVcfAsVRanges <- function(x, genome, param=ScanVcfParam(),
+                             use.names=FALSE, ...)
+{
+  if (!isTRUEorFALSE(use.names)) {
+    stop("'use.names' must be TRUE or FALSE")
+  }
+  as(readVcf(x, genome, param=param, row.names=use.names, ...), "VRanges")
 }
 
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 ### Writing to VCF (see methods-writeVcf.R)
 ###
 
+
+### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+### gVCF support
+###
+
+addTotalDepthRuns <- function(x, runs, genome) {
+  mins <- viewMins(runs)
+  gr <- as(ranges(runs), "GRanges")
+  ref <- getSeq(genome, resize(gr, 1L))
+  vr <- VRanges(seqnames(gr), ranges, ref, alt, totalDepth=mins)
+}
+
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 ### Filtering
 ###
@@ -485,11 +550,11 @@ setMethod("duplicated", "VRanges",
             if (!identical(incomparables, FALSE))
               stop("\"duplicated\" method for VRanges objects ",
                    "only accepts 'incomparables=FALSE'")
-            IRanges:::duplicatedIntegerQuads(as.factor(seqnames(x)),
-                                             as.factor(alt(x)),
-                                             start(x), width(x),
-                                             fromLast = fromLast,
-                                             method = method)
+            S4Vectors:::duplicatedIntegerQuads(as.factor(seqnames(x)),
+                                               as.factor(alt(x)),
+                                               start(x), width(x),
+                                               fromLast = fromLast,
+                                               method = method)
           })
 
 setMethod("match", c("VRanges", "VRanges"),
@@ -506,14 +571,15 @@ setMethod("match", c("VRanges", "VRanges"),
             merge(seqinfo(x), seqinfo(table))
             altLevels <- as.character(union(alt(x), alt(table)))
             x_seqnames <- GenomicRanges:::relevelSeqnamesForMatch(x, table)
-            IRanges:::matchIntegerQuads(x_seqnames,
-                                        factor(as.character(alt(x)), altLevels),
-                                        start(x), width(x),
-                                        as.factor(seqnames(table)),
-                                        factor(as.character(alt(table)),
-                                               altLevels),
-                                        start(table), width(table),
-                                        nomatch = nomatch, method = method)
+            S4Vectors:::matchIntegerQuads(x_seqnames,
+                                          factor(as.character(alt(x)),
+                                                 altLevels),
+                                          start(x), width(x),
+                                          as.factor(seqnames(table)),
+                                          factor(as.character(alt(table)),
+                                                 altLevels),
+                                          start(table), width(table),
+                                          nomatch = nomatch, method = method)
           })
 
 setMethod("%in%", c("VRanges", "TabixFile"),
