@@ -3,40 +3,42 @@
 ### =========================================================================
 
 setMethod("expand", "CollapsedVCF",
-    function(x, ...)
+    function(x, ..., row.names = FALSE)
     {
+        rd <- rowRanges(x, fixed=FALSE)
+        if (!row.names)
+            names(rd) <- NULL
+
         elt <- elementLengths(alt(x))
         if (all(elt == 1L)) {
             fxd <- fixed(x)
             fxd$ALT <- unlist(alt(x), use.names=FALSE)
-            if (!is.null(geno(x)$AD))
-              geno(x)$AD <- .expandAD(geno(x)$AD, nrow(x), ncol(x), elt)
-            return(VCF(rowData=.stripRowData(rowData(x)), colData=colData(x), 
-                       exptData=exptData(x), fixed=fxd, 
-                       info=.unlistAltInfo(x), geno=geno(x), 
-                       ..., collapsed=FALSE))
+            AD <- "AD" %in% names(geno(x))
+            if (AD)
+              geno(x)$AD <- .expandAD(geno(x)$AD, nrow(x), ncol(x))
+            return(VCF(rd, colData(x), exptData(x), fxd, .unlistAltInfo(x), 
+                       geno(x), ..., collapsed=FALSE))
         }
-        idx <- rep.int(seq_len(nrow(x)), elt)
+
+        ## info, fixed, geno
         hdr <- exptData(x)$header
-        ## info
+        idx <- rep.int(seq_len(nrow(x)), elt)
         iexp <- .expandInfo(x, hdr, elt, idx)
-        ## fixed
         fexp <- fixed(x)[idx, ]
         fexp$ALT <- unlist(alt(x), use.names=FALSE)
-        ## geno 
         gexp <- .expandGeno(x, hdr, elt, idx)
-        ## rowData
-        rdexp <- .stripRowData(rowData(x)[idx,])
-        ## exptData, colData untouched
-        VCF(rowData=rdexp, colData=colData(x), exptData=exptData(x),
-            fixed=fexp, info=iexp, geno=gexp, ..., collapsed=FALSE)
+
+        ## rowRanges
+        if (is.null(rd$paramRangeID)) {
+            rdexp <- rd[idx, ]
+            mcols(rdexp) <- NULL
+        } else rdexp <- rd[idx, "paramRangeID"]
+
+        VCF(rdexp, colData(x), exptData(x), fexp, iexp, gexp,
+            ..., collapsed=FALSE)
     }
 )
 
-.stripRowData <- function(x) {
-  x[,intersect(colnames(mcols(x)), "paramRangeID")]
-}
-
 .expandGeno <- function(x, hdr, elt, idx)
 {
     gvar <- geno(x)
@@ -47,8 +49,8 @@ setMethod("expand", "CollapsedVCF",
     isA <- ghdr$Number == "A"
     if (any(isA)) {
         gnms <- rownames(ghdr)[isA]
-        gelt <- sapply(gnms, function(i) 
-                    elt - elementLengths(gvar[[i]]))
+        gelt <- do.call(cbind, lapply(gnms, function(i) 
+                                      elt - elementLengths(gvar[[i]])))
         ## elementLengths same as ALT
         csums <- colSums(gelt) == 0L
         if (any(csums))
@@ -71,7 +73,8 @@ setMethod("expand", "CollapsedVCF",
         if (!is.list(AD)) {
             ## 'Number' is integer
             if (is(AD, "array")) {
-                if ((length(unique(elt)) != 1L) && (dim(AD)[3] != unique(elt))) {
+                if ((length(unique(elt)) != 1L) && 
+                    (dim(AD)[3] != unique(elt))) {
                     warning("'AD' was ignored: number of 'AD' values ",
                             "do not match REF + ALT")
                     isAD[isAD] <- FALSE 
@@ -81,7 +84,7 @@ setMethod("expand", "CollapsedVCF",
             }
         } else {
             ## 'Number' is '.'
-            gvar$AD <- .expandAD(AD, length(idx), ncol(x), elt)
+            gvar$AD <- .expandAD(AD, length(idx), ncol(x))
         }
     }
     isA <- names(gvar) %in% rownames(ghdr)[isA]
@@ -96,23 +99,19 @@ setMethod("expand", "CollapsedVCF",
     gvar
 }
 
-## returns an array of REF,ALT pairs
-.expandAD <- function(AD, idxlen, xcols, elt)
+## returns an array of REF,ALT and REF,<NON_REF> pairs
+## 'z' dimension of result is always 2
+.expandAD <- function(AD, idxlen, xcols)
 {
     if (is.list(AD)) {
-        emptyAD <- elementLengths(AD) == 0L
-        if (all(elementLengths(elt) == 1L)) {
-            AD[emptyAD] <- list(c(NA_integer_, NA_integer_))
-        } else {
-            emptypart <- PartitioningByWidth(elt[emptyAD])
-            na <- rep(NA_integer_, sum(width(emptypart)))
-            AD[emptyAD] <- as.list(relist(na, emptypart))
-        }
         adpart <- PartitioningByWidth(AD)
+        if (any(zeros <- width(adpart) == 0L)) { 
+            AD[zeros] <- list(rep(NA_integer_, 2L))
+            adpart <- PartitioningByWidth(AD)
+        }
+
         nalt <- width(adpart) - 1L
-        if (sum(nalt) != idxlen*xcols)
-          stop("length of AD does not match expanded index")
-        AD <- as.integer(unlist(AD, use.names=FALSE))
+        AD <- unlist(AD, use.names=FALSE)
         ref <- logical(length(AD))
         ref[start(adpart)] <- TRUE
         vec <- c(rep(AD[ref], nalt), AD[!ref])
@@ -128,16 +127,16 @@ setMethod("expand", "CollapsedVCF",
     ## no data 
     if (ncol(ivar) == 0L)
         return(DataFrame(row.names=seq_along(idx)))
-    inms <- rownames(info(hdr))[info(hdr)$Number == "A"]
+    inms <- names(ivar)[info(hdr)[names(ivar), "Number"] == "A"]
     if (length(inms) > 0L) {
-        ielt <- sapply(inms, function(i) 
-                    elt - elementLengths(ivar[[i]]))
+        ielt <- do.call(cbind, lapply(inms, function(i) 
+                                      elt - elementLengths(ivar[[i]])))
         ## elementLengths same as ALT
         csums <- colSums(ielt) == 0L
         if (any(csums))
             res <- IRanges:::.expandByColumnSet(ivar, inms[csums], TRUE)
         else
-            res <- ivar[idx, ] 
+            res <- ivar[idx, , drop=FALSE] 
         ## elementLengths shorter than ALT
         if (any(!csums)) {
             nms <- colnames(ivar) %in% names(csums)[!csums]
@@ -149,7 +148,7 @@ setMethod("expand", "CollapsedVCF",
         }
         res
     } else {
-        ivar[idx, ]
+        ivar[idx, , drop=FALSE]
     }
 }
 
@@ -167,9 +166,6 @@ setMethod("expand", "CollapsedVCF",
 }
 
 setMethod("expand", "ExpandedVCF",
-    function(x, ...)
-    {
-        x
-    }
+    function(x, ...) x
 )
 
