@@ -8,14 +8,14 @@
 ### 
 ### Each region has the following methods : 
 ### query %in% Ranges, VCF, GRanges
-### subject %in% TranscriptDb, GRangesList 
+### subject %in% TxDb, GRangesList 
 
 
 ### -------------------------------------------------------------------------
 ### methods applicable to all variant regions 
 ###
 
-setMethod("locateVariants", c("Ranges", "TranscriptDb", "VariantType"),
+setMethod("locateVariants", c("Ranges", "TxDb", "VariantType"),
     function(query, subject, region, ..., cache=new.env(parent=emptyenv()),
              ignore.strand=FALSE, asHits=FALSE)
 {
@@ -31,11 +31,11 @@ setMethod("locateVariants", c("Ranges", "GRangesList", "VariantType"),
         ignore.strand=ignore.strand, asHits=asHits)
 })
 
-setMethod("locateVariants", c("VCF", "TranscriptDb", "VariantType"),
+setMethod("locateVariants", c("VCF", "TxDb", "VariantType"),
     function(query, subject, region, ..., cache=new.env(parent=emptyenv()),
              ignore.strand=FALSE, asHits=FALSE)
 {
-    callGeneric(rowData(query), subject, region, ..., cache=cache,
+    callGeneric(rowRanges(query), subject, region, ..., cache=cache,
         ignore.strand=ignore.strand, asHits=asHits)
 })
 
@@ -43,7 +43,7 @@ setMethod("locateVariants", c("VCF", "GRangesList", "VariantType"),
     function(query, subject, region, ..., cache=new.env(parent=emptyenv()),
              ignore.strand=FALSE, asHits=FALSE)
 {
-    callGeneric(rowData(query), subject, region, ..., cache=cache,
+    callGeneric(rowRanges(query), subject, region, ..., cache=cache,
         ignore.strand=ignore.strand, asHits=asHits)
 })
 
@@ -51,14 +51,14 @@ setMethod("locateVariants", c("VCF", "GRangesList", "VariantType"),
 ### region = CodingVariants 
 ###
 
-setMethod("locateVariants", c("GRanges", "TranscriptDb", "CodingVariants"),
+setMethod("locateVariants", c("GRanges", "TxDb", "CodingVariants"),
     function(query, subject, region, ..., cache=new.env(parent=emptyenv()),
              ignore.strand=FALSE, asHits=FALSE)
     {
         if (!any(seqlevels(query) %in% seqlevels(subject)))
             return(.returnEmpty())
 
-        ## for width(ranges) == 0 : de-increment start to equal end value 
+       ## for width(ranges) == 0 : de-increment start to equal end value 
         if (any(insertion <- width(query) == 0))
             start(query)[insertion] <- start(query)[insertion] - 1
         if (!exists("cdsbytx", cache, inherits=FALSE))
@@ -85,7 +85,7 @@ setMethod("locateVariants", c("GRanges", "GRangesList", "CodingVariants"),
 ### region = IntronVariants 
 ###
 
-setMethod("locateVariants", c("GRanges", "TranscriptDb", "IntronVariants"),
+setMethod("locateVariants", c("GRanges", "TxDb", "IntronVariants"),
     function(query, subject, region, ..., cache=new.env(parent=emptyenv()),
              ignore.strand=FALSE, asHits=FALSE)
     {
@@ -119,7 +119,7 @@ setMethod("locateVariants", c("GRanges", "GRangesList", "IntronVariants"),
 ### region = ThreeUTRVariants 
 ###
 
-setMethod("locateVariants", c("GRanges", "TranscriptDb", "ThreeUTRVariants"),
+setMethod("locateVariants", c("GRanges", "TxDb", "ThreeUTRVariants"),
     function(query, subject, region, ..., cache=new.env(parent=emptyenv()),
              ignore.strand=FALSE, asHits=FALSE)
     {
@@ -129,7 +129,6 @@ setMethod("locateVariants", c("GRanges", "TranscriptDb", "ThreeUTRVariants"),
         ## for width(ranges) == 0 : de-increment start to equal end value 
         if (any(insertion <- width(query) == 0)) 
             start(query)[insertion] <- start(query)[insertion] - 1 
-
         if (!exists("threeUTRbytx", cache, inherits=FALSE))
             cache[["threeUTRbytx"]] <- threeUTRsByTranscript(subject)
         res <- callGeneric(query, cache[["threeUTRbytx"]], region, ...,
@@ -154,7 +153,7 @@ setMethod("locateVariants", c("GRanges", "GRangesList", "ThreeUTRVariants"),
 ### region = FiveUTRVariants 
 ###
 
-setMethod("locateVariants", c("GRanges", "TranscriptDb", "FiveUTRVariants"),
+setMethod("locateVariants", c("GRanges", "TxDb", "FiveUTRVariants"),
     function(query, subject, region, ..., cache=new.env(parent=emptyenv()),
              ignore.strand=FALSE, asHits=FALSE)
     {
@@ -164,7 +163,6 @@ setMethod("locateVariants", c("GRanges", "TranscriptDb", "FiveUTRVariants"),
         ## for width(ranges) == 0 : de-increment start to equal end value 
         if (any(insertion <- width(query) == 0))
             start(query)[insertion] <- start(query)[insertion] - 1
-
         if (!exists("fiveUTRbytx", cache, inherits=FALSE))
             cache[["fiveUTRbytx"]] <- fiveUTRsByTranscript(subject)
         res <- callGeneric(query, cache[["fiveUTRbytx"]], region, ...,
@@ -189,7 +187,7 @@ setMethod("locateVariants", c("GRanges", "GRangesList", "FiveUTRVariants"),
 ### region = IntergenicVariants 
 ###
 
-setMethod("locateVariants", c("GRanges", "TranscriptDb", 
+setMethod("locateVariants", c("GRanges", "TxDb", 
           "IntergenicVariants"),
     function(query, subject, region, ..., cache=new.env(parent=emptyenv()),
              ignore.strand=FALSE)
@@ -200,7 +198,6 @@ setMethod("locateVariants", c("GRanges", "TranscriptDb",
         ## for width(ranges) == 0 : de-increment start to equal end value 
         if (any(insertion <- width(query) == 0))
             start(query)[insertion] <- start(query)[insertion] - 1
-
         if (!exists("txbygene", cache, inherits=FALSE))
             cache[["txbygene"]] <- transcriptsBy(subject, "gene")
 
@@ -220,7 +217,7 @@ setMethod("locateVariants", c("GRanges", "GRangesList", "IntergenicVariants"),
 ## region = SpliceSiteVariants 
 ##
 
-setMethod("locateVariants", c("GRanges", "TranscriptDb", "SpliceSiteVariants"),
+setMethod("locateVariants", c("GRanges", "TxDb", "SpliceSiteVariants"),
     function(query, subject, region, ..., cache=new.env(parent=emptyenv()),
              ignore.strand=FALSE, asHits=FALSE)
     {
@@ -230,7 +227,6 @@ setMethod("locateVariants", c("GRanges", "TranscriptDb", "SpliceSiteVariants"),
         ## for width(ranges) == 0 : de-increment start to equal end value 
         if (any(insertion <- width(query) == 0))
             start(query)[insertion] <- start(query)[insertion] - 1
-
         if (!exists("intbytx", cache, inherits=FALSE))
             cache[["intbytx"]] <- intronsByTranscript(subject)
         res <- callGeneric(query, cache[["intbytx"]], region, ...,
@@ -254,7 +250,7 @@ setMethod("locateVariants", c("GRanges", "GRangesList", "SpliceSiteVariants"),
 ### region = PromoterVariants 
 ###
 
-setMethod("locateVariants", c("GRanges", "TranscriptDb", "PromoterVariants"),
+setMethod("locateVariants", c("GRanges", "TxDb", "PromoterVariants"),
     function(query, subject, region, ..., cache=new.env(parent=emptyenv()),
              ignore.strand=FALSE, asHits=FALSE)
     { 
@@ -264,7 +260,6 @@ setMethod("locateVariants", c("GRanges", "TranscriptDb", "PromoterVariants"),
         ## for width(ranges) == 0 : de-increment start to equal end value 
         if (any(insertion <- width(query) == 0)) 
             start(query)[insertion] <- start(query)[insertion] - 1 
-
         if (!exists("tx", cache, inherits=FALSE)) {
             tx <- transcripts(subject) 
             cache[["tx"]] <- splitAsList(tx, seq_len(length(tx))) 
@@ -302,17 +297,20 @@ setMethod("locateVariants", c("GRanges", "GRangesList", "PromoterVariants"),
                     ranges=IRanges(ranges(query)[queryid]),
                     strand=strand(pm)[subjectHits(fo)],
                     LOCATION=.location(length(queryid), "promoter"), 
+                    LOCSTART=NA_integer_,
+                    LOCEND=NA_integer_,
                     QUERYID=queryid,
                     TXID=as.integer(txid[subjectHits(fo)]),
-                    CDSID=NA_integer_,
+                    CDSID=IntegerList(integer(0)),
                     GENEID=NA_character_,
                     PRECEDEID=CharacterList(character(0)),
                     FOLLOWID=CharacterList(character(0)))
         } else {
             res <- GRanges()
-            values(res) <- DataFrame(LOCATION=.location(), QUERYID=integer(),
-                                     TXID=integer(), CDSID=integer(),
-                                     GENEID=character(),
+            values(res) <- DataFrame(LOCATION=.location(), 
+                                     LOCSTART=integer(), LOCEND=integer(),
+                                     QUERYID=integer(), TXID=integer(), 
+                                     CDSID=IntegerList(), GENEID=character(),
                                      PRECEDEID=CharacterList(),
                                      FOLLOWID=CharacterList())
             res
@@ -324,7 +322,7 @@ setMethod("locateVariants", c("GRanges", "GRangesList", "PromoterVariants"),
 ### region = AllVariants 
 ###
 
-setMethod("locateVariants", c("GRanges", "TranscriptDb", "AllVariants"),
+setMethod("locateVariants", c("GRanges", "TxDb", "AllVariants"),
     function(query, subject, region, ..., cache=new.env(parent=emptyenv()),
              ignore.strand=FALSE)
     {
@@ -372,7 +370,7 @@ setMethod("locateVariants", c("GRanges", "TranscriptDb", "AllVariants"),
 
         ans <- c(coding, intron, fiveUTR, threeUTR, splice, promoter, intergenic)
         meta <- values(ans)
-        ans[order(meta$QUERYID, meta$TXID, meta$CDSID, meta$GENEID), ]
+        ans[order(meta$QUERYID, meta$TXID, meta$GENEID), ]
     }
 )
 
@@ -384,9 +382,11 @@ setMethod("locateVariants", c("GRanges", "TranscriptDb", "AllVariants"),
 {
     warning("none of seqlevels(query) match seqlevels(subject)")
     res <- GRanges()
-    values(res) <- DataFrame(LOCATION=.location(), QUERYID=integer(), 
-                             TXID=integer(), CDSID=integer(), 
-                             GENEID=character(), PRECEDEID=CharacterList(), 
+    values(res) <- DataFrame(LOCATION=.location(),
+                             LOCSTART=integer(), LOCEND=integer(),
+                             QUERYID=integer(), TXID=integer(), 
+                             CDSID=IntegerList(), GENEID=character(), 
+                             PRECEDEID=CharacterList(), 
                              FOLLOWID=CharacterList())
     res
 }
@@ -417,24 +417,26 @@ setMethod("locateVariants", c("GRanges", "TranscriptDb", "AllVariants"),
         return(.consolidateHits(fo, length(query), length(subject),
             elementLengths(subject))) 
     if (length(fo) > 0) {
-        txid <- rep(names(subject), elementLengths(subject))
-        queryid <- queryHits(fo)
-        subject_strand <- c(strand(int_start), strand(int_end))
-        GRanges(seqnames=seqnames(query)[queryid],
-                ranges=IRanges(ranges(query)[queryid]),
-                strand=subject_strand[subjectHits(fo)],
-                LOCATION=.location(length(queryid), "spliceSite"),
-                QUERYID=queryid,
-                TXID=as.integer(txid[subjectHits(fo)]),
-                CDSID=NA_integer_,
+        df <- unique(data.frame(queryid=queryHits(fo),
+                     subjectid=togroup(subject)[subjectHits(fo)]))
+        GRanges(seqnames=seqnames(query)[df$queryid],
+                ranges=IRanges(ranges(query)[df$queryid]),
+                strand=unlist(strand(subject), use.names=FALSE)[df$subjectid],
+                LOCATION=.location(length(df$queryid), "spliceSite"),
+                LOCSTART=NA_integer_, LOCEND=NA_integer_,
+                QUERYID=df$queryid,
+                TXID=as.integer(names(subject)[df$subjectid]),
+                CDSID=IntegerList(integer(0)),
                 GENEID=NA_character_,
                 PRECEDEID=CharacterList(character(0)),
                 FOLLOWID=CharacterList(character(0)))
     } else {
         res <- GRanges()
-        values(res) <- DataFrame(LOCATION=.location(), QUERYID=integer(),
-                                 TXID=integer(), CDSID=integer(), 
-                                 GENEID=character(), PRECEDEID=CharacterList(),
+        values(res) <- DataFrame(LOCATION=.location(), 
+                                 LOCSTART=integer(), LOCEND=integer(),
+                                 QUERYID=integer(), TXID=integer(), 
+                                 CDSID=IntegerList(), GENEID=character(), 
+                                 PRECEDEID=CharacterList(),
                                  FOLLOWID=CharacterList())
         res
     }
@@ -445,11 +447,7 @@ setMethod("locateVariants", c("GRanges", "TranscriptDb", "AllVariants"),
     txid <- rep(seq_len(slen), elen)
     dat <- cbind(queryHits(hits), as.integer(txid[subjectHits(hits)]))
     unq <- dat[!duplicated(dat),]
-    return(new("Hits", 
-           queryHits=unq[,1], 
-           subjectHits=unq[,2],
-           queryLength=qlen, 
-           subjectLength=slen)) 
+    return(Hits(unq[,1], unq[,2], qlen, slen))
 }
 
 .intergenic <- function(query, subject, region, ignore.strand, ...)
@@ -457,12 +455,18 @@ setMethod("locateVariants", c("GRanges", "TranscriptDb", "AllVariants"),
     s_range <- range(subject) ## avoid duplicates 
     co <- unname(countOverlaps(query, s_range, type="any", 
                                ignore.strand=ignore.strand))
-    intergenic <- co == 0
+
+    ## variants that don't hit a gene feature AND
+    ## zero-width ranges return '0'
+    has_width <- width(query) != 0L
+    intergenic <- co == 0 & has_width
     if (all(!intergenic | (length(subject) == 0L))) {
         res <- GRanges()
-        values(res) <- DataFrame(LOCATION=.location(), QUERYID=integer(), 
-                                 TXID=integer(), CDSID=integer(), 
-                                 GENEID=character(), PRECEDEID=CharacterList(), 
+        values(res) <- DataFrame(LOCATION=.location(), 
+                                 LOCSTART=integer(), LOCEND=integer(),
+                                 QUERYID=integer(), TXID=integer(), 
+                                 CDSID=IntegerList(), GENEID=character(), 
+                                 PRECEDEID=CharacterList(), 
                                  FOLLOWID=CharacterList())
         res
     } else {
@@ -473,13 +477,13 @@ setMethod("locateVariants", c("GRanges", "TranscriptDb", "AllVariants"),
         ## ID all genes that fall in upstream / downstream range.
         ## upstream == follow:
         f_range <- .shiftRangeUpDown(q_range, upstream(region), TRUE)
-        f_fo <- findOverlaps(f_range, s_unlist)
+        f_fo <- findOverlaps(f_range, s_unlist, ignore.strand=ignore.strand)
         f_factor <- factor(queryHits(f_fo), seq_len(queryLength(f_fo)))
         f_genes <- unname(splitAsList(s_genes[subjectHits(f_fo)], f_factor))
  
         ## downstream == precede:
         p_range <- .shiftRangeUpDown(q_range, downstream(region), FALSE)
-        p_fo <- findOverlaps(p_range, s_unlist)
+        p_fo <- findOverlaps(p_range, s_unlist, ignore.strand=ignore.strand)
         p_factor <- factor(queryHits(p_fo), seq_len(queryLength(p_fo)))
         p_genes <- unname(splitAsList(s_genes[subjectHits(p_fo)], p_factor))
 
@@ -487,8 +491,9 @@ setMethod("locateVariants", c("GRanges", "TranscriptDb", "AllVariants"),
             strand(q_range) <- "*"
         values(q_range) <- 
             DataFrame(LOCATION=.location(length(q_range), "intergenic"),
+                      LOCSTART=NA_integer_, LOCEND=NA_integer_,
                       QUERYID=which(intergenic), TXID=NA_integer_, 
-                      CDSID=NA_integer_, GENEID=NA_character_, 
+                      CDSID=IntegerList(integer(0)), GENEID=NA_character_, 
                       PRECEDEID=p_genes, FOLLOWID=f_genes)
         q_range 
     }
@@ -523,40 +528,59 @@ setMethod("locateVariants", c("GRanges", "TranscriptDb", "AllVariants"),
 
 .makeResult <- function(query, subject, vtype, ignore.strand, asHits)
 {
-    if (asHits) {
+    if (asHits)
         return(findOverlaps(query, subject, type="within", 
-            ignore.strand=ignore.strand))
-    } else {
-        usub <- unlist(subject, use.names=FALSE)
-        fo <- findOverlaps(query, usub, type="within", 
-            ignore.strand=ignore.strand)
-    }
-    if (length(fo) > 0) {
-        queryid <- queryHits(fo)
-        txid <- NA_integer_
-        cdsid <- NA_integer_
-        if (!is.null(tx <- rep(names(subject), elementLengths(subject))))
-            txid <- tx 
-        if (vtype == "coding")
-            if(!is.null(cds <- values(usub)[["cds_id"]][subjectHits(fo)]))
-                cdsid <- cds
+                            ignore.strand=ignore.strand))
 
-        GRanges(seqnames=seqnames(query)[queryid],
-                ranges=IRanges(ranges(query)[queryid]),
-                strand=strand(usub)[subjectHits(fo)],
-                LOCATION=.location(length(queryid), vtype),
-                QUERYID=queryid,
-                TXID=as.integer(txid[subjectHits(fo)]),
+    map <- mapToTranscripts(unname(query), subject, 
+                            ignore.strand=ignore.strand)
+    if (length(map) > 0) {
+        xHits <- map$xHits
+        txHits <- map$transcriptsHits
+        if (!is.null(tx <- names(subject)[txHits]))
+            txid <- tx
+        else
+            txid <- NA_integer_
+        ## FIXME: cdsid is expensive
+        cdsid <- IntegerList(integer(0))
+        if (vtype == "coding") {
+           usub <- unlist(subject) ## names needed for mapping
+            map2 <- mapToTranscripts(unname(query)[xHits], usub,
+                                     ignore.strand=ignore.strand)
+            cds <- mcols(usub)$cds_id[map2$transcriptsHits]
+            if (length(cds)) {
+                cdslst <- unique(splitAsList(cds, map2$xHits))
+                cdsid <- cdslst
+            }
+        }
+
+        ss <- runValue(strand(subject)[txHits])
+        if (any(elementLengths(ss) > 1L)) {
+            warning("'subject' has multiple strands per list element; ",
+                    "setting strand to '*'")
+            sstrand <- Rle("*", length(txHits))
+        }
+        sstrand <- unlist(ss, use.names=FALSE)
+        GRanges(seqnames=seqnames(query)[xHits],
+                ranges=IRanges(ranges(query)[xHits]),
+                strand=sstrand,
+                LOCATION=.location(length(xHits), vtype),
+                LOCSTART=start(map),
+                LOCEND=end(map),
+                QUERYID=xHits,
+                TXID=txid,
                 CDSID=cdsid,
                 GENEID=NA_character_,
                 PRECEDEID=CharacterList(character(0)),
                 FOLLOWID=CharacterList(character(0)))
     } else {
         res <- GRanges()
-        values(res) <- DataFrame(LOCATION=.location(), QUERYID=integer(),
-                                 TXID=integer(), CDSID=integer(), 
-                                 GENEID=character(), PRECEDEID=CharacterList(),
-                                 FOLLOWID=CharacterList())
+        mcols(res) <- DataFrame(LOCATION=.location(),
+                                LOCSTART=integer(), LOCEND=integer(),
+                                QUERYID=integer(), TXID=integer(), 
+                                CDSID=IntegerList(), GENEID=character(), 
+                                PRECEDEID=CharacterList(),
+                                FOLLOWID=CharacterList())
         res
     }
 }
